from typing import List, Optional

from pydantic import BaseModel

from cbpa.schemas.account import Account
from cbpa.schemas.alert import DiscordAlert
from cbpa.schemas.api import CoinbaseProAPI
from cbpa.schemas.buy import Buy


class Config(BaseModel):
    api: CoinbaseProAPI
    account: Account
    discord: Optional[DiscordAlert]
    buys: List[Buy]



from cbpa.schemas.buy import Buy


def test_pair() -> None:
    buy = Buy(send_currency="USD", send_amount=1, receive_currency="BTC")
    assert buy.pair() == "BTC-USD"
    assert buy.pair() != "USD-BTC"



from pydantic import BaseModel
from pydantic.types import PositiveInt

from cbpa.schemas.currency import CCC, FCC


class Buy(BaseModel):
    send_currency: FCC = FCC.USD
    send_amount: PositiveInt
    receive_currency: CCC

    def pair(self) -> str:
        return f"{self.receive_currency}-{self.send_currency}"


class BuyResponse(BaseModel):
    message: str



import json
import math
import time
from typing import Union

import coinbasepro

from cbpa.logger import logger
from cbpa.schemas.account import AddFundsResponse
from cbpa.schemas.config import Config
from cbpa.schemas.currency import CCC, FCC
from cbpa.services.discord import DiscordService


class AccountService:
    def __init__(
        self, config: Config, coinbasepro_client: coinbasepro.AuthenticatedClient
    ) -> None:
        self.coinbasepro_client = coinbasepro_client
        self.config = config
        self.discord = DiscordService()

    def get_balance_for_currency(self, currency: Union[CCC, FCC]) -> int:
        """get_balance_for_currency

        Assuming there will only every be one account for a currency identifier.
        This account is not the same as a payment method.
        Think of it as just the allocation of a currency.
        """
        return [
            math.floor(account["balance"])
            for account in self.coinbasepro_client.get_accounts()
            if account["currency"] == currency.value
        ][0]

    def get_primary_buy_payment_method_id(self, fiat: FCC) -> str:
        return [
            primary_method["id"]
            for primary_method in self.coinbasepro_client.get_payment_methods()
            if primary_method["primary_buy"]
            and primary_method["currency"] == fiat.value
        ][0]

    def add_funds(
        self,
        buy_total: int,
        current_funds: int,
        max_fund: int,
        fiat: FCC,
    ) -> AddFundsResponse:
        if buy_total > max_fund:
            message = (
                f"Total cost is {buy_total} {fiat} but you "
                f"have your limit set to {max_fund} {fiat}. "
                "Unable to complete purchase. "
                "Update your settings appropriately to make changes if necessary."
            )
            self.discord.send_alert(config=self.config, message=message)
            return AddFundsResponse(status="Error", message=message)
        else:
            fund_amount = buy_total - current_funds
            if current_funds > 1:
                fund_amount = max_fund
            fund_message = (
                f"Your balance is {current_funds} {fiat}. "
                f"A deposit of {fund_amount} {fiat} will be made "
                "using your selected payment account."
            )
            logger.info(fund_message)
            self.discord.send_alert(config=self.config, message=fund_message)
            payment_id = self.get_primary_buy_payment_method_id(fiat=fiat)
            if payment_id is None:
                return AddFundsResponse(
                    status="Error", message="Could not determine payment account id."
                )
            else:
                deposit = self.coinbasepro_client.deposit(
                    amount=fund_amount,
                    currency=fiat.value,
                    payment_method_id=payment_id,
                )
                logger.info("Sleeping for 10 seconds while the deposit completes.")
                time.sleep(10)
                return AddFundsResponse(
                    status="Success",
                    message=json.dumps(deposit, sort_keys=True, default=str),
                )



import time
import traceback
from typing import Dict

import coinbasepro

from cbpa.logger import logger
from cbpa.schemas.buy import Buy
from cbpa.schemas.config import Config
from cbpa.schemas.currency import FCC
from cbpa.services.account import AccountService
from cbpa.services.discord import DiscordService


class BuyService:
    MARKET_SIDE = "buy"

    def __init__(
        self,
        config: Config,
        coinbasepro_client: coinbasepro.AuthenticatedClient,
        account_service: AccountService,
    ) -> None:
        self.coinbasepro_client = coinbasepro_client
        self.discord_service = DiscordService()
        self.config = config
        self.account_service = account_service

    def place_market_order(self, buy: Buy, fiat: FCC) -> None:
        pair = buy.pair()
        amount = buy.send_amount
        place_order_message = f"💸 Placing market order of {amount} {fiat} for {pair}."
        logger.info(place_order_message)
        response = self.coinbasepro_client.place_market_order(
            product_id=pair, side=self.MARKET_SIDE, funds=amount
        )
        logger.info("Order placed! Getting order details.")
        self.get_placed_order_details(buy=buy, response=response)

    def get_placed_order_details(self, buy: Buy, response: Dict) -> None:
        order_id = response["id"]
        try:
            order_details = self.coinbasepro_client.get_order(order_id=order_id)
            order_filled_size = order_details["filled_size"]
            purchase_success_message = (
                f"👏 Successfully purchased {order_filled_size} of {buy.pair()}."
            )
            logger.info(purchase_success_message)
            self.discord_service.send_alert(
                config=self.config, message=purchase_success_message
            )
        except coinbasepro.exceptions.CoinbaseAPIError as e:
            logger.error(
                f"Coinbase returned an API error: {e}. "
                f"Order {order_id} was not created just yet. "
                "Retrying order detail retrieval in 2 seconds."
            )
            time.sleep(2)
            self.get_placed_order_details(buy=buy, response=response)

    def run(self, done: Dict) -> None:
        try:
            for buy in self.config.buys:
                if not done[buy.receive_currency.value]:
                    buy_total = buy.send_amount
                    current_funds = self.account_service.get_balance_for_currency(
                        self.config.account.fiat_currency
                    )
                    if current_funds >= buy_total:
                        self.place_market_order(buy, self.config.account.fiat_currency)
                        done[buy.receive_currency.value] = True
                    elif current_funds < buy_total:
                        response = self.account_service.add_funds(
                            buy_total=buy_total,
                            current_funds=current_funds,
                            max_fund=self.config.account.auto_funding_limit,
                            fiat=self.config.account.fiat_currency,
                        )
                        if response.status == "Error":
                            logger.error(response.message)
                            self.discord_service.send_alert(
                                config=self.config, message=response.message
                            )
                        elif response.status == "Success":
                            self.place_market_order(
                                buy, self.config.account.fiat_currency
                            )
                            done[buy.receive_currency.value] = True
                        else:
                            logger.info(
                                f"Unhandled response status {response.status}."
                                " Moving on."
                            )
            if not all(done.values()):
                self.run(done=done)
        except Exception:
            logger.error("Unhandled general exception occurred.")
            logger.error(traceback.format_exc())
            self.run(done=done)



import os
from datetime import datetime
from typing import Optional

import coinbasepro
import typer
import uvicorn
import yaml
from fastapi import FastAPI
from google.cloud import secretmanager

from cbpa import __version__
from cbpa.logger import logger
from cbpa.schemas.buy import BuyResponse
from cbpa.schemas.config import Config
from cbpa.schemas.health import HealthcheckResponse
from cbpa.services.account import AccountService
from cbpa.services.buy import BuyService
from cbpa.services.config import ConfigService
from cbpa.services.discord import DiscordService

os.environ["TZ"] = "UTC"


def create_config(filepath: str) -> Config:
    config_service = ConfigService()
    logger.info(f"🤓 Reading config file at {filepath}.")
    config = config_service.load_config(filepath=filepath)
    logger.info(f"👏 Successfully parsed {filepath}.")
    return config


def retrieve_config_from_secret_manager(
    project_id: str, secret_id: str, version_id: str = "latest"
) -> Config:
    client = secretmanager.SecretManagerServiceClient()
    name = f"projects/{project_id}/secrets/{secret_id}/versions/{version_id}"
    response = client.access_secret_version(request={"name": name})
    payload = response.payload.data.decode("UTF-8")
    return Config(**yaml.safe_load(payload))


def create_coinbasepro_auth_client(config: Config) -> coinbasepro.AuthenticatedClient:
    logger.info("📡 Connecting to Coinbase Pro.")
    client = coinbasepro.AuthenticatedClient(
        key=config.api.key.get_secret_value(),
        secret=config.api.secret.get_secret_value(),
        passphrase=config.api.passphrase.get_secret_value(),
        api_url=config.api.url,
    )
    logger.info("👏 Successfully connected to Coinbase Pro.")
    return client


def create_account_service(
    config: Config, coinbasepro_client: coinbasepro.AuthenticatedClient
) -> AccountService:
    logger.info("🏦 Creating account service.")
    account_service = AccountService(
        config=config,
        coinbasepro_client=coinbasepro_client,
    )
    logger.info("👏 Successfully created account service.")
    return account_service


def create_buy_service(
    config: Config,
    coinbasepro_client: coinbasepro.AuthenticatedClient,
    account_service: AccountService,
) -> BuyService:
    logger.info("💸 Creating buy service.")
    buy_service = BuyService(
        config=config,
        coinbasepro_client=coinbasepro_client,
        account_service=account_service,
    )
    logger.info("👏 Successfully created buy service.")
    return buy_service


def _main(config: Config) -> None:
    start = datetime.now()
    discord_service = DiscordService()
    start_message = f"🤖 Starting cbpa ({start.isoformat()})"
    logger.info(start_message)
    discord_service.send_alert(config=config, message=start_message)
    coinbasepro_client = create_coinbasepro_auth_client(config=config)
    account_service = create_account_service(
        config=config, coinbasepro_client=coinbasepro_client
    )
    buy_service = create_buy_service(
        config=config,
        coinbasepro_client=coinbasepro_client,
        account_service=account_service,
    )
    done = {buy.receive_currency.value: False for buy in config.buys}

    buy_service.run(done=done)

    end = datetime.now()
    duration = end - start
    end_message = f"🤖 cbpa completed! Ran for {duration.total_seconds()} seconds."
    logger.info(end_message)
    discord_service.send_alert(config=config, message=end_message)


#
#   create the api
#
api = FastAPI(
    title="Coinbase Pro Automation API",
    description="Automate buys for your favourite cryptocurrencies.",
    version=__version__,
)


#
#   create api routes
#
@api.get("/healthcheck", response_model=HealthcheckResponse, tags=["health"])
def healthcheck() -> HealthcheckResponse:
    message = "Dollar cost averaging."
    logger.debug(message)
    return HealthcheckResponse(
        api_version=__version__,
        message=message,
        time=datetime.now(),
    )


@api.post("/buy", response_model=BuyResponse, tags=["buy"])
def buy() -> BuyResponse:
    logger.info("Buy API request initiated.")
    config = retrieve_config_from_secret_manager(
        project_id=os.environ["PROJECT_ID"], secret_id=os.environ["SECRET_ID"]
    )
    _main(config=config)
    logger.info("Buy API request completed.")
    return BuyResponse(message="Buy API request completed.")


#
#   create the cli
#
typer_app_name = "Coinbase Pro Automation"
app = typer.Typer(name=typer_app_name)


#
#   create cli commands
#
@app.command("version", help="prints the version")
def _version() -> None:
    typer.echo(__version__)


@app.command("run", help="executes buy orders listed in a config file")
def _run(
    filepath: str = typer.Option(
        ...,
        "-f",
        "--file",
        help="filepath to the yaml config",
    )
) -> None:
    config = create_config(filepath=filepath)
    _main(config=config)


@app.command("server", help="run an api server to handle automated buys")
def _server(
    port: Optional[str] = typer.Option(
        None, "--port", "-p", help="the port to run uvicon on"
    ),
    host: str = typer.Option(
        "0.0.0.0", "--host", "-h", help="the port to run uvicon on"
    ),
) -> None:
    if port is None:
        port = os.getenv("PORT", "8002")
    assert port is not None
    uvicorn.run(api, port=int(port), host=host)



import os
from typing import Generator

import pytest
from typer.testing import CliRunner

os.environ["TZ"] = "UTC"


@pytest.fixture(scope="function")
def cli_runner() -> Generator:
    yield CliRunner()



import logging
import os
import socket
import time


def create_logger() -> logging.Logger:
    tz = time.strftime("%z")
    logging.basicConfig(
        format=(
            f"[%(asctime)s.%(msecs)03d {tz}] "
            f"[%(process)s] [{socket.gethostname()}] [%(pathname)s L%(lineno)d] "
            "[%(levelname)s] %(message)s"
        ),
        level=os.environ.get("LOGLEVEL", "INFO").upper(),
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    logger = logging.getLogger(__name__)
    return logger


logger = create_logger()



from pydantic import BaseModel, SecretStr, StrictStr


class CoinbaseProAPI(BaseModel):
    key: SecretStr
    secret: SecretStr
    passphrase: SecretStr
    url: StrictStr = "https://api.pro.coinbase.com"



import yaml

from cbpa.schemas.config import Config


class ConfigService:
    def __init__(self) -> None:
        pass

    def load_config(self, filepath: str) -> Config:
        with open(filepath, "r") as f:
            content = yaml.safe_load(f)
        return Config(**content)



import uuid
from typing import Dict, List


class MockCoinbaseClient:
    def get_accounts(self) -> List[Dict]:
        return [{"balance": 100, "currency": "USD"}]

    def get_payment_methods(self) -> List[Dict]:
        return [
            {
                "id": "900229be-9eb0-4e54-96b2-6a7ccf2a7c51",
                "primary_buy": True,
                "currency": "USD",
            }
        ]

    def deposit(self, amount: float, currency: str, payment_method_id: str) -> None:
        return None

    def place_market_order(self, product_id: str, side: str, funds: float) -> Dict:
        return {"id": str(uuid.uuid4())}

    def get_order(self, order_id: str) -> Dict:
        return {"filled_size": 2}



from discord import RequestsWebhookAdapter, Webhook

from cbpa.schemas.config import Config


class DiscordService:
    def __init__(self) -> None:
        pass

    def send_alert(self, config: Config, message: str) -> None:
        if config.discord:
            webhook = Webhook.from_url(
                config.discord.webhook,
                adapter=RequestsWebhookAdapter(),
            )
            webhook.send(message)



from enum import Enum, unique

from pydantic import BaseModel, PositiveInt, StrictStr

from cbpa.schemas.currency import FCC


@unique
class AddFundsStatus(str, Enum):
    Success = "Success"
    Error = "Error"


class Account(BaseModel):
    auto_funding_limit: PositiveInt
    fiat_currency: FCC = FCC.USD


class AddFundsResponse(BaseModel):
    status: AddFundsStatus
    message: StrictStr



import os
from unittest import mock

from typer.testing import CliRunner

from cbpa.main import app
from tests.mocks import MockCoinbaseClient


def test_timezone() -> None:
    assert os.environ["TZ"] == "UTC"


@mock.patch(
    "cbpa.main.create_coinbasepro_auth_client",
    return_value=MockCoinbaseClient(),
)
@mock.patch(
    "cbpa.services.discord.DiscordService.send_alert",
    return_value=None,
)
def test_run_with_alerts(
    mock_discord_service: mock.MagicMock,
    mock_coinbase_client: mock.MagicMock,
    cli_runner: CliRunner,
) -> None:
    cli_runner.invoke(app, ["run", "-f", "./examples/cbpa.yaml"])


@mock.patch(
    "cbpa.main.create_coinbasepro_auth_client",
    return_value=MockCoinbaseClient(),
)
@mock.patch(
    "cbpa.services.discord.DiscordService.send_alert",
    return_value=None,
)
def test_run_without_alerts(
    mock_discord_service: mock.MagicMock,
    mock_coinbase_client: mock.MagicMock,
    cli_runner: CliRunner,
) -> None:
    cli_runner.invoke(app, ["run", "-f", "./examples/cbpa-no-alerts.yaml"])


@mock.patch(
    "cbpa.main.create_coinbasepro_auth_client",
    return_value=MockCoinbaseClient(),
)
@mock.patch(
    "cbpa.services.discord.DiscordService.send_alert",
    return_value=None,
)
def test_run_trigger_over_limit(
    mock_discord_service: mock.MagicMock,
    mock_coinbase_client: mock.MagicMock,
    cli_runner: CliRunner,
) -> None:
    cli_runner.invoke(app, ["run", "-f", "./examples/cbpa-trigger-over-limit.yaml"])


@mock.patch(
    "cbpa.main.create_coinbasepro_auth_client",
    return_value=MockCoinbaseClient(),
)
@mock.patch(
    "cbpa.services.discord.DiscordService.send_alert",
    return_value=None,
)
def test_run_trigger_deposit(
    mock_discord_service: mock.MagicMock,
    mock_coinbase_client: mock.MagicMock,
    cli_runner: CliRunner,
) -> None:
    cli_runner.invoke(app, ["run", "-f", "./examples/cbpa-trigger-deposit.yaml"])



from enum import Enum, unique


@unique
class CCC(str, Enum):
    """CCC stands for _C_rypto_C_urrency _C_ode"""

    BTC = "BTC"
    ETH = "ETH"
    DOGE = "DOGE"
    SHIB = "SHIB"
    MANA = "MANA"
    AVAX = "AVAX"
    SOL = "SOL"
    XTZ = "XTZ"


@unique
class FCC(str, Enum):
    """FCC stands for _F_iat _C_urrency _C_ode"""

    USD = "USD"
